# -*- coding: utf-8 -*-
import fnmatch


class ResumableFile(object):
    def __init__(self, storage, kwargs):
        self.storage = storage
        self.kwargs = kwargs
        self.chunk_suffix = "_part_"

    @property
    def chunk_exists(self):
        """Checks if the requested chunk exists."""
        name = "%s%s%s" % (self.filename,
                           self.chunk_suffix,
                           self.kwargs.get('resumableChunkNumber').zfill(4))
        if not self.storage.exists(name):
            return False
        chunk_size = int(self.kwargs.get('resumableCurrentChunkSize'))
        return self.storage.size(name) == chunk_size

    def chunk_names(self):
        """Iterates over all stored chunks and yields their names."""
        file_names = sorted(self.storage.listdir('')[1])
        pattern = '%s%s*' % (self.filename, self.chunk_suffix)
        for name in file_names:
            if fnmatch.fnmatch(name, pattern):
                yield name

    def chunks(self):
        """Yield the contents of every chunk, FileSystemStorage.save compatible
        """
        for name in self.chunk_names():
            yield self.storage.open(name).read()

    def delete_chunks(self):
        [self.storage.delete(chunk) for chunk in self.chunk_names()]

    @property
    def filename(self):
        """Gets the filename."""
        filename = self.kwargs.get('resumableFilename')
        if '/' in filename:
            raise Exception('Invalid filename')
        return "%s_%s" % (
            self.kwargs.get('resumableTotalSize'),
            filename
        )

    @property
    def is_complete(self):
        """Checks if all chunks are allready stored."""
        if self.storage.exists(self.filename):
            return True
        return int(self.kwargs.get('resumableTotalSize')) == self.size

    def process_chunk(self, file):
        if not self.chunk_exists:
            self.storage.save('%s%s%s' % (
                self.filename,
                self.chunk_suffix,
                self.kwargs.get('resumableChunkNumber').zfill(4)
            ), file)

    @property
    def size(self):
        """Gets chunks size."""
        size = 0
        for chunk in self.chunk_names():
            size += self.storage.size(chunk)
        return size



# -*- coding: utf-8 -*-
from django.forms.fields import FileField
from django.core.exceptions import ValidationError
from django.utils.translation import ugettext_lazy as _

from .widgets import ResumableFileInput


class ResumableFileField(FileField):
    default_error_messages = dict(FileField.default_error_messages, **{
            'invalid_mime': _(u'Invalid file type')
    })
    widget = ResumableFileInput

    def __init__(self, *args, **kwargs):
        self.allowed_mimes = kwargs.pop('allowed_mimes', None)
        upload_url = kwargs.pop('upload_url', None)
        chunks_dir = kwargs.pop('chunks_dir', None)
        super(ResumableFileField, self).__init__(*args, **kwargs)
        self.upload_url = upload_url
        self.chunks_dir = chunks_dir

    @property
    def chunks_dir(self):
        if not hasattr(self.widget, 'chunks_dir'):
            raise Exception("You must set a `chunk_dir`.")
        return self.widget.storage

    @chunks_dir.setter
    def chunks_dir(self, chunks_dir):
        self.widget.chunks_dir = chunks_dir

    def clean(self, data, initial):
        f = super(ResumableFileField, self).clean(data, initial)
        if self.allowed_mimes is not None and \
                f.content_type not in self.allowed_mimes:
            raise ValidationError(self.error_messages['invalid_mime'])
        return f

    @property
    def upload_url(self):
        if 'data-upload-url' not in self.widget.attrs:
            raise Exception("You must set the upload url.")

        return self.widget.attrs['data-upload-url']

    @upload_url.setter
    def upload_url(self, url):
        self.widget.attrs['data-upload-url'] = url



# -*- coding: utf-8 -*-
import os

from urllib.parse import urlencode

from django.conf import settings
from django.test import TestCase
from django.core.files.base import ContentFile
from django.core.files.uploadedfile import UploadedFile
from django.core.files.storage import FileSystemStorage
from django.core.urlresolvers import reverse
from django.core.exceptions import ValidationError

from resumable.files import ResumableFile

from .app import ResumableForm


TESTS_ROOT = os.path.dirname(__file__)


FIXTURES_ROOT = os.path.join(TESTS_ROOT, 'fixtures', 'files')


CHUNKS_ROOT = os.path.join(FIXTURES_ROOT, 'chunks')


seagull = {
    'resumableTotalSize': '147292',
    'resumableFilename': 'seagull.ogg',
    'resumableChunkNumber': '8',
}


craw = {
    'resumableTotalSize': '49028',
    'resumableFilename': 'craw.ogg',
    'resumableChunkNumber': '4',
    'resumableCurrentChunkSize': 18308,
}


class BaseTestCase(TestCase):
    def setUp(self):
        test_storage = FileSystemStorage(
            location=getattr(settings, 'FILE_UPLOAD_TEMP_DIR'))
        fixtures_storage = FileSystemStorage(location=CHUNKS_ROOT)

        for filename in fixtures_storage.listdir('.')[1]:
            test_storage.save(
                filename,
                fixtures_storage.open(filename)
            )
        self.seagull = ResumableFile(test_storage, seagull)
        self.craw = ResumableFile(test_storage, craw)
        self.storage = test_storage

    def tearDown(self):
        for filename in self.storage.listdir('.')[1]:
            self.storage.delete(filename)


class ResumableFileFieldTest(BaseTestCase):
    def test_clean_invalid_mime(self):
        form = ResumableForm()
        with self.assertRaises(ValidationError):
            form.fields.get('file').clean(None, UploadedFile(
                file=None,
                name="text.txt",
                content_type="text/plain"
            ))

    def test_clean_valid_mime(self):
        form = ResumableForm()
        f = UploadedFile(
            file=None,
            name="sound.ogg",
            content_type="audio/ogg"
        )
        self.assertEqual(f, form.fields.get('file').clean(None, f))

    def test_form_upload_file(self):
        r = self.client.post(reverse('form'), {'file': open(os.path.join(
            FIXTURES_ROOT, '147292_seagull.ogg'), 'rb')})
        self.assertEqual(r.status_code, 302)


class ResumableFileTest(BaseTestCase):
    def test_chunks_partial(self):
        iterations = 0
        for chunk in self.seagull.chunks():
            iterations += 1
        self.assertEqual(iterations, 7)

    def test_chunks_complete(self):
        data = b''
        for chunk in self.craw.chunks():
            data += chunk
        self.assertEqual(len(data), 49028)

    def test_chunk_exists_existing(self):
        self.assertTrue(self.craw.chunk_exists)

    def test_chunk_exists_missing(self):
        self.assertFalse(self.seagull.chunk_exists)

    def test_filename(self):
        self.assertEqual(self.seagull.filename, '147292_seagull.ogg')

    def test_is_complete_complete(self):
        self.assertTrue(self.craw.is_complete)

    def test_is_complete_partial(self):
        self.assertFalse(self.seagull.is_complete)

    def test_process_chunk(self):
        self.assertFalse(self.seagull.chunk_exists)
        chunk = ContentFile('content')
        self.seagull.kwargs['resumableCurrentChunkSize'] = chunk.size
        self.seagull.process_chunk(chunk)
        self.assertTrue(self.seagull.chunk_exists)

    def test_size_complete(self):
        self.assertEqual(self.craw.size, 49028)

    def test_size_partial(self):
        self.assertEqual(self.seagull.size, 71680)


class ResumableUploadViewTest(BaseTestCase):
    def test_get_existing(self):
        url = '%s?%s' % (reverse('upload'), urlencode(craw))
        r = self.client.get(url)
        self.assertEqual(r.status_code, 200)

    def test_get_missing(self):
        url = '%s?%s' % (reverse('upload'), urlencode(seagull))
        r = self.client.get(url)
        self.assertEqual(r.status_code, 404)

    def test_post_missing(self):
        self.assertFalse(self.seagull.chunk_exists)
        path = os.path.join(CHUNKS_ROOT, 'chunk')
        size = os.path.getsize(path)
        chunk = open(path)
        params = dict(seagull, **{
            'file': chunk,
            'resumableCurrentChunkSize': size,
        })
        self.seagull.kwargs['resumableCurrentChunkSize'] = size
        self.client.post(reverse('upload'), params)
        self.assertTrue(self.seagull.chunk_exists)



# -*- coding: utf-8 -*-
import magic
import re

from django.forms.widgets import FileInput
from django.core.files.storage import FileSystemStorage
from django.core.files.uploadedfile import UploadedFile
from django.template import loader


class ResumableFileInput(FileInput):
    template_name = 'resumable/file_input.html'

    def __init__(self, *args, **kwargs):
        super(ResumableFileInput, self).__init__(*args, **kwargs)
        self.filepath = None
        self.filename = None

    def filename_input_name(self, name):
        """Returns the name of the input[type=hidden] that contains the
        uploaded file path
        """
        return "%s-path" % name

    def guess_type(self, path):
        mime = magic.Magic(mime=True)
        return mime.from_file(path)

    def value_from_datadict(self, data, files, name):
        filepath = data.get(self.filename_input_name(name))
        storage = self.storage
        if filepath is not None and len(filepath) > 0 \
                and storage.exists(filepath):
            file = storage.open(filepath)
            size = storage.size(filepath)
            self.filepath = filepath
            self.filename = re.sub('^%s_' % str(size), '', filepath)
            return UploadedFile(
                file=file,
                name=self.filename,
                content_type=self.guess_type(file.name),
                size=size
            )
        return files.get(name, None)

    def render(self, name, value, attrs=None, **kwargs):
        attrs = self.build_attrs(attrs)

        if 'required' in attrs:
            del attrs['required']

        context = {
            'filename': self.filename,
            'filename_input_value': self.filepath,
            'filename_input_name': self.filename_input_name(name),
            'file_input_name': name,
            'attrs': attrs,
        }
        return loader.render_to_string(self.template_name, context)

    @property
    def storage(self):
        return FileSystemStorage(location=self.chunks_dir)



# -*- coding: utf-8 -*-
from django.conf import settings
from django.conf.urls import url
from django.conf.urls.static import static
from django.views.generic.edit import FormView
from django.forms import Form
from django.core.urlresolvers import reverse
from django.contrib.staticfiles.urls import staticfiles_urlpatterns

from resumable.views import ResumableUploadView
from resumable.fields import ResumableFileField


class ResumableForm(Form):
    file = ResumableFileField(
        allowed_mimes=("audio/ogg",),
        upload_url=lambda: reverse('upload'),
        chunks_dir=getattr(settings, 'FILE_UPLOAD_TEMP_DIR')
    )


class TestFormView(FormView):
    form_class = ResumableForm
    template_name = 'form.html'

    @property
    def success_url(self):
        return reverse('form')


urlpatterns = staticfiles_urlpatterns()


urlpatterns = [
    url('^$', TestFormView.as_view(), name='form'),
    url('^upload/$', ResumableUploadView.as_view(), name='upload')
] + static(settings.STATIC_URL, document_root=settings.STATIC_ROOT)



# -*- coding: utf-8 -*-
from django.conf import settings

from django.views.generic import View
from django.http import HttpResponse
from django.core.exceptions import ImproperlyConfigured
from django.core.files.storage import FileSystemStorage

from resumable.files import ResumableFile


class ResumableUploadView(View):
    def get(self, *args, **kwargs):
        """Checks if chunk has allready been sended.
        """
        r = ResumableFile(self.storage, self.request.GET)
        if not (r.chunk_exists or r.is_complete):
            return HttpResponse('chunk not found', status=404)
        return HttpResponse('chunk already exists')

    def post(self, *args, **kwargs):
        """Saves chunks then checks if the file is complete.
        """
        chunk = self.request.FILES.get('file')
        r = ResumableFile(self.storage, self.request.POST)
        if r.chunk_exists:
            return HttpResponse('chunk already exists')
        r.process_chunk(chunk)
        if r.is_complete:
            self.process_file(r.filename, r)
            r.delete_chunks()
        return HttpResponse()

    def process_file(self, filename, file):
        """Process the complete file.
        """
        self.storage.save(filename, file)

    @property
    def chunks_dir(self):
        chunks_dir = getattr(settings, 'FILE_UPLOAD_TEMP_DIR', None)
        if not chunks_dir:
            raise ImproperlyConfigured(
                'You must set settings.FILE_UPLOAD_TEMP_DIR')
        return chunks_dir

    @property
    def storage(self):
        return FileSystemStorage(location=self.chunks_dir)



# -*- coding: utf-8 -*-
from tempfile import mkdtemp


DEBUG = True

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'resumable.db',
    },
}

INSTALLED_APPS = [
    'django.contrib.staticfiles',
    'resumable',
    'resumable.tests.app',
]

STATIC_URL = '/static/'

STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
)

SECRET_KEY = 'secret'

ROOT_URLCONF = 'resumable.tests.app'

FILE_UPLOAD_TEMP_DIR = mkdtemp()

MIDDLEWARE_CLASSES = [
    'django.middleware.csrf.CsrfViewMiddleware',
]

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'APP_DIRS': True,
    },
]



# -*- coding: utf-8 -*-
from distutils.core import setup
from setuptools import find_packages


setup(
    name='django-resumable',
    version='0.2.0',
    author=u'jean-philippe serafin',
    author_email='serafinjp@gmail.com',
    packages=find_packages(),
    include_package_data=True,
    url='https://github.com/jeanphix/django-resumable',
    license='MIT licence',
    description='Django resumable uploads',
    long_description=open('README.rst').read(),
    install_requires=[
        'Django>=1.10',
        'python-magic',
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python',
    ],
    zip_safe=False,
)
